from __future__ import annotations

import click

from . import make_config, make_gh_action, make_readme, make_template


@click.group()
def make():
    """Generate, make, produce, print different things."""


make.add_command(make_readme.readme)
make.add_command(make_template.template)
make.add_command(make_gh_action.gh_action)
make.add_command(make_config.config)



from __future__ import annotations

import logging

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.search import commit, rebuild

log = logging.getLogger(__name__)


class CascadeOrganizationUpdatesPlugin(p.SingletonPlugin):
    p.implements(p.IActions)

    def get_actions(self):
        return {
            "organization_update": organization_update,
        }


@tk.chained_action
def organization_update(next_, context, data_dict):
    """Reindex all datasets inside the organization after any update."""
    result = next_(context, data_dict)
    tk.enqueue_job(reindex_organization, [result["id"]])
    return result


def reindex_organization(id_or_name: str):
    """Rebuild search index for all datasets inside the organization."""
    org = model.Group.get(id_or_name)

    if not org:
        log.warning("Organization with ID or name %s not found", id_or_name)
        return

    query = model.Session.query(model.Package.id).filter_by(owner_org=org.id)

    rebuild(package_ids=(p.id for p in query), force=True)
    commit()



from __future__ import annotations

import dataclasses
from collections.abc import Collection
from typing import Any, Generic, TypeVar

T = TypeVar("T")


@dataclasses.dataclass
class Node(Generic[T]):
    value: T
    leaves: Collection[Node[T]] = ()
    parent: Node | None = None
    data: dict[str, Any] = dataclasses.field(default_factory=dict)

    def __iter__(self):
        return iter(self.leaves)

    def __len__(self):
        return len(self.leaves)

    def __eq__(self, other: Any):
        if not isinstance(other, Node):
            return False

        return self.value == other.value and self.leaves == other.leaves



#             *
#             / \
#            /___\
#           ( o o )            * *
#           )  L  (           /   * *
#   ________()(-)()________  /     * * *
# E\| _____ )()()() ______ |/B     * * *
#   |/      ()()()(       \|      * * * *
#           | )() |
#           /     \
#          / *  *  \
#         /   *  *  \
#        / *_  *  _  \
###############################################################################
#                  I solemnly swear that I am up to no good,                  #
###############################################################################

import logging

import ckan.plugins.toolkit as tk

log = logging.getLogger(__name__)


def conjure_fast_group_activities():
    log.info("ieiunium sicut ventus")

    def ___group_activity_perfomance_patch(group_id, include_hidden_activity=False):
        from ckan import model

        group = model.Group.get(group_id)
        if not group:
            # Return a query with no results.
            return model.Session.query(model.Activity).filter(text("0=1"))  # noqa

        q = model.Session.query(model.Activity)
        group_activity = q.filter(model.Activity.object_id == group_id)
        packages_sq = (
            model.Session.query(model.Package.id)
            .filter_by(owner_org=group_id, private=False)
            .subquery()
        )

        member_activity = model.Session.query(model.Activity).filter(
            model.Activity.object_id.in_(packages_sq),
        )

        if not include_hidden_activity:
            group_activity = _filter_activitites_from_users(group_activity)  # noqa
            member_activity = _filter_activitites_from_users(member_activity)  # noqa
        return _activities_union_all(group_activity, member_activity)  # noqa

    if tk.check_ckan_version("2.10"):
        from ckanext.activity.model.activity import _group_activity_query
    else:
        from ckan.model.activity import _group_activity_query

    _group_activity_query.__code__ = ___group_activity_perfomance_patch.__code__


def transfigure_xloaded_file(func):
    """Proces a file before uploading it to datastore.

    Accepts callable, that receives path to original file and resource
    ID. Callable can override file in order to provide a valid CSV that can be
    ingested into datastore.

    """
    from ckanext.xloader import loader

    log.info("quae non sunt ut simplex")

    _o = loader.load_csv

    def _wrapper(csv_filepath, resource_id, mimetype="text/csv", logger=None):
        new_path = func(csv_filepath, resource_id)
        return _o(new_path, resource_id, mimetype, logger)

    loader.load_csv = _wrapper



from __future__ import annotations

import logging

from werkzeug.utils import import_string

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model

SQLAlchemy = import_string("flask_sqlalchemy:SQLAlchemy", True)

_EngineDebuggingSignalEvents = import_string(
    "flask_sqlalchemy:_EngineDebuggingSignalEvents",
    True,
)
record_queries = import_string("flask_sqlalchemy.record_queries", True)


log = logging.getLogger(__name__)


class FdtSqlalchemyPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurable)
    p.implements(p.IMiddleware, inherit=True)

    def make_middleware(self, app, config):
        if not SQLAlchemy:
            version = "3.0" if tk.check_ckan_version("2.11.0") else "2.5"
            log.error(
                "Flask-SQLAlchemy is not installed. "
                "Run `pip install flask-sqlalchemy~=%s` and restart the application",
                version,
            )
            return app

        app.config.setdefault("SQLALCHEMY_TRACK_MODIFICATIONS", False)
        app.config["SQLALCHEMY_DATABASE_URI"] = config["sqlalchemy.url"]
        SQLAlchemy().init_app(app)

        return app

    def configure(self, _config):
        if _EngineDebuggingSignalEvents:
            _EngineDebuggingSignalEvents(model.meta.engine, "ckan").register()

        if record_queries:
            record_queries._listen(model.meta.engine)



from __future__ import annotations

import logging
import os
import tempfile

import requests

import ckan.plugins as p
from ckan.lib.uploader import get_resource_uploader

log = logging.getLogger(__name__)

DEFAULT_DOWNLOAD_TIMEOUT = 2


class StaticPath:
    def __init__(self, path: str | None):
        self.path = path

    def __bool__(self):
        return bool(self.path)

    def __enter__(self):
        return self.path

    def __exit__(self, type_, value, traceback):
        pass


class RemovablePath(StaticPath):
    def __exit__(self, type_, value, traceback):
        if self.path:
            os.remove(self.path)


def path_to_resource(res, max_size: int = 0) -> StaticPath:
    """Returns a filepath for a resource that will be indexed"""
    res_id = res["id"]
    res_url = res["url"]

    if res["url_type"] == "upload":
        uploader = get_resource_uploader(res)

        # TODO temporary workaround for ckanext-cloudstorage support
        if p.plugin_loaded("cloudstorage"):
            url = uploader.get_url_from_filename(res_id, res_url)
            filepath = _download_remote_file(res_id, url, max_size)
            return RemovablePath(filepath)

        path = uploader.get_path(res_id)
        if not os.path.exists(path):
            log.warning('Resource "%s" refers to unexisting path "%s"', res, path)
            return StaticPath(None)

        return StaticPath(path)

    if max_size > 0:
        filepath = _download_remote_file(res_id, res_url, max_size)
        return RemovablePath(filepath)

    return StaticPath(None)


def _download_remote_file(res_id: str, url: str, max_size: int) -> str | None:
    """
    Downloads remote resource and save it as temporary file
    Returns path to this file
    """

    try:
        resp = requests.get(
            url,
            timeout=DEFAULT_DOWNLOAD_TIMEOUT,
            stream=True,
            headers={"user-agent": "python/toolbelt"},
        )
    except requests.RequestException as e:
        log.warning(
            "Unable to make GET request for resource %s with url <%s>: %s",
            res_id,
            url,
            e,
        )
        return None

    if not resp.ok:
        log.warning(
            "Unsuccessful GET request for resource %s with url <%s>. \
            Status code: %s",
            res_id,
            url,
            resp.status_code,
        )
        return None

    try:
        size = int(resp.headers.get("content-length", 0))
    except ValueError:
        log.warning("Incorrect Content-length header from url <%s>", url)
        return None

    if not size:
        log.debug("Cannot determine size")
        return None

    if size > max_size:
        log.debug("File exceeds allowed size(%d): %d", max_size, size)
        return None

    dest = tempfile.NamedTemporaryFile(delete=False)
    try:
        with dest:
            for chunk in resp.iter_content(1024 * 64):
                dest.write(chunk)
    except requests.RequestException:
        log.exception("Cannot index remote resource %s with url <%s>", res_id, url)
        os.remove(dest.name)
        return None
    return dest.name



import click

from .dev import dev
from .make import make


@click.group(short_help="Variety of useful commands.")
def toolbelt():
    """Variety of useful commands."""


toolbelt.add_command(make)
toolbelt.add_command(dev)



from __future__ import annotations

import ckan.plugins as p
import ckan.plugins.toolkit as tk


class FdtScrollPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurer, inherit=True)

    def update_config(self, config):
        if tk.asbool(tk.config.get("debug")):
            tk.add_template_directory(config, "templates")



from __future__ import annotations

from sqlalchemy.orm.scoping import ScopedSession
from typing_extensions import TypeAlias, TypedDict

import ckan.model as model_

Model: TypeAlias = "model_"
AlchemySession = ScopedSession


class SelectOption(TypedDict):
    text: str | None
    value: str



from __future__ import annotations

import copy
from typing import Any, Callable, Generic, TypeVar

from typing_extensions import Self

TFunc = TypeVar("TFunc", bound=Callable[..., Any])


class Collector(Generic[TFunc]):
    collection: dict[str, TFunc]

    def __init__(self, prefix: str = "", separator: str = "_"):
        self.collection = {}

        if prefix:
            self.prefix = prefix + separator

        else:
            self.prefix = ""

    def split(self) -> tuple[Self, Callable[[], dict[str, TFunc]]]:
        return self, self.get_collection

    def __call__(self, func_or_name: str | TFunc):
        name: str

        def adder(func: TFunc):
            self.collection[name] = func
            return func

        if isinstance(func_or_name, str):
            name = func_or_name
            return adder

        name = self.prefix + func_or_name.__name__
        return adder(func_or_name)

    def get_collection(self) -> dict[str, TFunc]:
        return copy.copy(self.collection)
